/*   ###########################################################################################################################################
 *   ###   Home automation version 2017.3                                                                                                    ###
 *   ###                                                                                                                                     ###
 *   ###   Program is made for an Arduino Mega with ESP8266 WIFI module, a RF 433 module, an eight relay module and an OLED screen (IC2)     ###
 *   ###   and a pushbutton.                                                                                                                 ###
 *   ###                                                                                                                                     ###
 *   ###   The website of this device uses extra files that are nessecary for functioning.                                                   ###
 *   ###                                                                                                                                     ###
 *   ###   More info can be found on http://www.2muchtime.org                                                                                ###
 *   ###########################################################################################################################################
 */


/*   ###########################################################################################################################################
 *   ###   HOME AUTOMATION BASIC INFO                                                                                                        ###
 *   ###########################################################################################################################################
 */

const String ssid        = "#####";                           // <<==This is your wifi SSID
const String password    = "##########";                   // <<==This is the password which belongs to the network with the SSID above
const String websiteName = "Home Automation";                 // <<==This will be the title of the generated website

/*   ###########################################################################################################################################
 *   ###   CONNECTED DEVIVES AND THEIR DESTINATIONS                                                                                          ###
 *   ###########################################################################################################################################
 */
//How many devices?
const int devices = 11;

//Names of the devices. Used for website and OLED.
const String deviceNames[] = {
  "Television", "Moodlight", "Light near window", "TV light", "Moodlight bed", "Bedside table light", "Desk light",   "Light at backside", "General", "Light above table", "Front door" 
};

//Tab on website
const int deviceTabs[] = {
  1, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4 
};

//Destinations below 10 are relays, 11 and higher are radio signals
const int destinations[] = {
  1, 2, 3, 4, 11, 12, 7, 13, 14, 15, 16
};

//Available codes to turn on (first is destination 11, and so on)
const long int rfCodesTurnOn[] ={
  12063919, 12063917, 12063915, 3700175, 3700173, 8429393
};

//Available codes to turn off (first is destination 11, and so on)
const long int rfCodesTurnOff[] = {
  12063918, 12063916, 12063914, 3700174, 3700172, 8429393
};

/*   ###########################################################################################################################################
 *   ###    WEBSITE                                                                                                                          ###
 *   ###########################################################################################################################################
 */
const char website_01[] PROGMEM = "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n<!DOCTYPE html><head><link href=\"http://2muchtime.org/projects/ha/ha.css\" rel=\"stylesheet\" /><script src=\"https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js\"></script></head><body ><div id=\"c\" style=\"display:hidden\">";
const char website_02[] PROGMEM = "</div></body><script src=\"http://2muchtime.org/projects/ha/ha.js\"></script></html>";
const char* const string_table[] PROGMEM = { website_01, website_02};


/*   ###########################################################################################################################################
 *   ###    SETUP                                                                                                                            ###
 *   ###########################################################################################################################################
 */

//Setup RC Switch 
#include <RCSwitch.h>
RCSwitch mySwitch = RCSwitch();
#define pinTX         53
#define pulseLength   308
#define bitLength     24
#define protocol      1

//Setup OLED screen
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);
#define LOGO16_GLCD_HEIGHT 16 
#define LOGO16_GLCD_WIDTH  16

//Bitmap of wifi logo
static const unsigned char PROGMEM logo16_glcd_bmp[] =
{ B00000000, B00000000, B01111000, B00000000, B00011111, B00000000, B00000011, B10000000,
  B00000000, B11100000, B01100000, B01110000, B01111100, B00110000, B00001110, B00011000,
  B00000011, B00001100, B01000001, B10001100, B01110000, B11000100, B00011000, B11000110,
  B00001100, B01000010, B01100100, B01100010, B01100110, B01100010, B00000000, B00000000 };

//Global vars
const String charsAllowed = "0123456789abcdefOK.:\"";       //<-- These are the only characters which will recorded from response from ESP8266
String lastMessage = "";
long int displayStartUp = millis();
String myIP = "";
char response[80];
boolean connection = false;

void setup() {
  //Serial.begin(9600);
  Serial1.begin(115200);

  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);                 //<-- The address(0x0C) can be different on other OLED screens. 
  pinMode(50, INPUT);                                        //<-- Button

  //Define relay and set all channels to off
  for (int i=24; i<38; i = i + 2){
    pinMode (i,OUTPUT);
    digitalWrite (i, HIGH);
  }

  //433 RF unit
  mySwitch.enableTransmit(pinTX);
  mySwitch.setPulseLength(pulseLength);
  mySwitch.setProtocol(protocol);
}

/*   ###########################################################################################################################################
 *   ###    MAIN PROGRAM                                                                                                                     ###
 *   ###########################################################################################################################################
 */


void loop() {
  while (connection == false){
    lastMessage = "Connecting to WIFI";
    OLED();
    useESP("AT+RST\r\n", 2000, true);
    useESP("AT+CWMODE=1\r\n", 2500, false);
    useESP("AT+CWJAP=\"" + ssid + "\",\"" + password + "\"\r\n", 6000, false);
    useESP("AT+CIPSTATUS\r\n", 2500, false);
    useESP("AT+CIPMUX=1\r\n", 2500, false);
    useESP("AT+CIPSERVER=1,80\r\n", 2500, false); 
    if (useESP("AT+CIFSR\r\n", 1500, false) == false) {
      //Second attempt, seems to work often
      useESP("AT+CIFSR\r\n", 1500, false);
    }
    if (connection == false){
      lastMessage = "Retry connecting in 30s.";
      OLED();      
      long int time = millis();    
      while( (time + 30000) > millis()) {
        if (digitalRead(50) == HIGH) {
          displayStartUp = millis();        //<--On a buttonpress, the OLED lights up for a minute
          OLED();
        }
        delay (50);
      }    
    } else {
      lastMessage = "Connected";
    }
  }

  while (connection == true) {

    OLED();
    String fromBuffer = "";
    boolean startRecording = false;
    boolean stopRecording = false;
    String connectionId;
    String urlQuery;
    long int startRecordingTime;
    long int lastCommunication = millis();
    int c;
    int tmr = 5000;
    
    while (stopRecording == false && connection == true){
        c = -1;
        while (c < 0){
          c = Serial1.read();
          if (digitalRead(50) == HIGH) {
            displayStartUp = millis();
          }
          if (tmr > 4999) {
            OLED();
            tmr = 0;
          }
          tmr++;
          if ((millis() - lastCommunication) > 300000) {
            //Because it takes time, the system only checks one time every 5 minutes for connection.
            useESP("AT+CIFSR\r\n", 1500, false);
            if (myIP == "0.0.0.0"){
              lastMessage = "Connection lost";
            }
          }
        }
       
        fromBuffer += char(c);
        if (startRecording == false) {
          if (fromBuffer.indexOf("+IPD") != -1) { 
            startRecording = true;
            startRecordingTime = millis();
            fromBuffer = fromBuffer.substring(fromBuffer.indexOf("+IPD"));
          }
          if (fromBuffer.length() > 20){
            fromBuffer = fromBuffer.substring(16);
          }
        } else {
          if (fromBuffer.indexOf("HTTP") > 0) {
            stopRecording = true;
            connectionId = fromBuffer.substring(fromBuffer.indexOf(",") + 1);
            connectionId = connectionId.substring(0,connectionId.indexOf(","));
            urlQuery = fromBuffer.substring(fromBuffer.indexOf(" ") + 1);
            urlQuery = urlQuery.substring(0 , urlQuery.indexOf(" "));
           }
           if (millis() - startRecordingTime > 5000){
              //When there is no end marker (HTTP) recognised after 5 seconds, the program returns to its original state: waiting for +IPD
              startRecording = false;
              startRecordingTime = 0;
              fromBuffer = fromBuffer.substring(4); //Removes the old +IPD from the bufferstring
           }
      }
      if (startRecording == false) {
        delay(10);
      }
    }
    if (stopRecording == true){
      if (urlQuery == "/") {
        lastMessage = "Serve website";
        OLED();
        serveWebsite(connectionId);
      }
      for (int i = 1; i <= devices; i++){
        if ((urlQuery.indexOf("/?s" + String(i) + "=0") != -1)){
          turnSwitch(i - 1, 1); //Turn switch on
        }
         if ((urlQuery.indexOf("/?s" + String(i) + "=1") != -1)){
          turnSwitch(i - 1, 0); //Turn switch off
        }
      }
    }
    useESP ("AT+CIPCLOSE=5\r\n", 500, false);
  } 
}

/*   ###########################################################################################################################################
 *   ###    SUB ROUTINES                                                                                                                     ###
 *   ###########################################################################################################################################
 */


boolean useESP(String command, const int timeout, boolean memflush) {
  //This part communicates with the ESP.
  //command = the command you want to give (don't forget the \r\n)! (string)
  //timeout = the time to wait if an expected response is not present. (millis)
  //memflush = tries to empty the buffer / cleas all response (true or false)
  delay(200);
  int responsePointer = 0;
  boolean responseReady = false;
  boolean commandOK = false;
  Serial1.print(command); 
  long int time = millis();
  while(responseReady == false) {
    char c = Serial1.read();
    if (c != -1){
      if (charsAllowed.indexOf(c) != -1 ){
        response[responsePointer] = c;
        responsePointer++;
      }        
    }
    if (millis() - time > timeout){
      responseReady = true;
    }
    if ((responsePointer >= 2)&& (response[responsePointer - 2] == 'O') && (response[responsePointer - 1] == 'K')){
      commandOK = true;
      responseReady = true;
    }
  }
  if (memflush){
    time = millis();
    while (millis()-time < 400){
      while(Serial1.available() ) { 
        char c = Serial1.read();
        time = millis();
      }
    }
  }
  //On a succesfull CIFSR command, an IP address must be extracted from the result
  if (command == "AT+CIFSR\r\n" && commandOK == true){
    myIP = "";
    int toggleLocator = 1;
    for (int i = 0; i < responsePointer; i++){
      if (response[i] == '"'){
        toggleLocator ++;
      } else {
        if (toggleLocator == 2){
          myIP += response[i];
        }
      }
    }
    if (myIP == "0.0.0.0"){
      //No IP is no connection
      connection = false;
    } else {
      connection = true;
    }
  }
  if (commandOK){
    return true;
    } else {
    return false;
  }
}


void OLED(){
  //Let the screen show the desired info. Standard it shows the (global) var myIP and the last received order.

  if ((displayStartUp + 60000) > millis()){
    
    display.clearDisplay(); 
    display.setTextSize(1);
    display.setTextColor(WHITE);
  
    if (connection == true){
    
      display.drawBitmap(1, 1,  logo16_glcd_bmp, 16, 16, 1); 
      display.setCursor(20,4);
      display.println(ssid);   
      display.setCursor(0,52);
      display.println("IP: " + myIP);       
    }
    
    display.setCursor(0,24);
    display.println(lastMessage);
    display.display();
    
  } else {
    
    display.clearDisplay(); 
    display.display();
    
  }
}


void serveWebsite(String conId){
  //Is able to serve the website to a given connection ID (int)
  //Does not use the function useESP, as I do not want to hand strings of this size over to another function.
  String sitevars = websiteName;
  int curtab = 0;
  
  for (int i = 0; i < devices; i++){
    if (deviceTabs[i] != curtab){
      sitevars += ",*";
      curtab ++;
    }
    sitevars += "," + deviceNames[i];
  }

  int messageLength = strlen(website_01) + strlen(website_02) + sitevars.length();
  Serial1.println("AT+CIPSEND=" + String(conId) + "," + String(messageLength) + "\r\n");
  delay(1000);
  Serial1.println((__FlashStringHelper*)pgm_read_word(string_table + 0));
  Serial1.println(sitevars);
  Serial1.println((__FlashStringHelper*)pgm_read_word(string_table + 1));
  delay(2000);
  Serial1.println("AT+CIPCLOSE=" + conId + "\r\n");

}


void turnSwitch (int switchno, int state){
  //Is used to switch a relay or broadcast a signal to turn things on and off
  if (destinations[switchno] < 10){
    if (state == 0){
      lastMessage = "Turned on:" + deviceNames[switchno];
      digitalWrite (22 + (destinations[switchno] * 2), LOW);
    } else {
      lastMessage = "Turned off:" + deviceNames[switchno];
      digitalWrite (22 + (destinations[switchno] * 2), HIGH);
    }
  } else {
    //Some RF codes are not on / off, but a pulse (doors for example)
    //If the code for rtCodesTurnOn and Off are the same, the text "Turn on/off" will be changed by "activate"
    lastMessage = (rfCodesTurnOn[destinations[switchno] - 11] == rfCodesTurnOff[destinations[switchno] - 11] ? "Activate:" : (state == 0 ? "Turned on:" : "Turned off:")) + deviceNames[switchno];
    if (state == 0){
       mySwitch.send(rfCodesTurnOn[destinations[switchno] - 11], bitLength);
    } else {
      mySwitch.send(rfCodesTurnOff[destinations[switchno] - 11], bitLength);
    }
  }
  OLED();  
}

